use super::prelude::*;
use futures::{StreamExt, TryStreamExt};
use rocket_db_pools::mongodb::options::FindOptions;
use util::{Ordering, PartyError};

api_routes!(
    add_user,
    get_user,
    get_all_users,
    delete_user,
    set_score,
    get_score
);

/// # User
///
/// A user that represents a person participating in the LAN party
#[derive(Clone, Debug, FromForm, Serialize, Deserialize, JsonSchema)]
#[serde(crate = "rocket::serde")]
pub struct User {
    /// Name of the user
    name: String,
    /// Score of the user
    score: i64,
    /// Unique identifier of the user
    id: String,
}

/// # Create new user with the give name
///
/// Returns the created user
#[openapi(tag = "User")]
#[post("/", data = "<name>")]
pub async fn add_user(
    _api_key: ApiKey,
    db: Connection<Db>,
    name: Json<&str>,
) -> Result<status::Created<Json<User>>, PartyError> {
    let user = User {
        id: uuid::Uuid::new_v4().to_string(),
        score: 0,
        name: name.to_string(),
    };

    db.users().insert_one(&user, None).await?;

    Ok(status::Created::new("/").body(Json(user)))
}

/// # Delete user by id
#[openapi(tag = "User")]
#[delete("/<id>")]
pub async fn delete_user(
    _api_key: ApiKey,
    db: Connection<Db>,
    id: String,
) -> Result<Status, PartyError> {
    db.users().delete_one(doc! { "id": id }, None).await?;

    Ok(Status::Ok)
}

/// # Get user by id
///
/// Returns a single user by id
#[openapi(tag = "User")]
#[get("/<id>")]
pub async fn get_user(
    _api_key: ApiKey,
    db: Connection<Db>,
    id: String,
) -> Result<Json<User>, PartyError> {
    let user = db
        .users()
        .find_one(doc! { "id": &id }, None)
        .await?
        .ok_or(PartyError::UserNotFound(id))?;
    Ok(Json(user))
}

/// # UserSort
///
/// Field used to sort users
#[derive(Clone, Debug, FromFormField, JsonSchema)]
#[serde(rename_all = "snake_case")]
pub enum UserSort {
    #[field(value = "score")]
    Score,
    #[field(value = "name")]
    Name,
    #[field(value = "id")]
    Id,
}

impl ToString for UserSort {
    fn to_string(&self) -> String {
        match self {
            Self::Score => "score",
            Self::Name => "name",
            Self::Id => "id",
        }
        .into()
    }
}

/// # Get all users
///
/// Returns an array of all users sorted by the given sort field and ordering
#[openapi(tag = "User")]
#[get("/?<sort>&<order>")]
pub async fn get_all_users(
    _api_key: ApiKey,
    db: Connection<Db>,
    sort: Option<UserSort>,
    order: Option<Ordering>,
) -> Result<Json<Vec<User>>, PartyError> {
    let options = if let Some(sort) = sort {
        let order = order.unwrap_or(Ordering::Asc);
        let mut doc = Document::new();
        doc.insert(
            sort.to_string(),
            match order {
                Ordering::Asc => 1,
                Ordering::Desc => -1,
            },
        );
        FindOptions::builder().sort(doc).build()
    } else {
        FindOptions::builder().build()
    };

    let users = db
        .users()
        .find(doc! {}, Some(options))
        .await?
        .try_collect()
        .await?;

    Ok(Json(users))
}

/// # Set score of user by id
#[openapi(tag = "User")]
#[post("/<id>/score", data = "<score>")]
pub async fn set_score(
    _api_key: ApiKey,
    db: Connection<Db>,
    id: String,
    score: Json<i64>,
) -> Result<Status, PartyError> {
    db.users()
        .update_one(
            doc! { "id": id },
            doc! { "$set": { "score": *score } },
            None,
        )
        .await?;

    Ok(Status::Ok)
}

/// # Get score of user by id
///
/// Returns the score of a single user by id
#[openapi(tag = "User")]
#[get("/<id>/score")]
pub async fn get_score(
    _api_key: ApiKey,
    db: Connection<Db>,
    id: String,
) -> Result<Json<i64>, PartyError> {
    Ok(Json(
        db.users()
            .find_one(doc! { "id": &id }, None)
            .await?
            .ok_or(PartyError::UserNotFound(id))?
            .score,
    ))
}
